/* GDB to RDI widget
   Copyright 2005 CodeSourcery, LLC
   Licence to be decided.  */
/* TODO: This whole file is rather spaghetti-like.  It needs breaking up, or
   at least dividing into logical chunks.  */
/* TODO: Most of the gdb error codes "ENN" are random guesses, and probably
   wrong.  */

#include <config.h>

#include <stdarg.h>
#include <stdio.h>
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#ifdef HAVE_DLFCN_H
#include <dlfcn.h>
#endif
#define RDI_VERSION 151
#include <rdi.h>
#include <sordi.h>
#include <rdi_hif.h>
#include <rdi_sdm.h>
#include <toolconf.h>

#ifdef HAVE_WINSOCK2_H
#include <winsock2.h>
#define SOCKET_OK(s) (s != INVALID_SOCKET)
typedef int socklen_t;

#else
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>

typedef int SOCKET;
#define SOCKET_OK(s) (s >= 0)
#define closesocket(s) close(s)
#define SD_SEND SHUT_WR

#endif

/* GDB remote protocol register numbers.  */
#define NUM_GDB_G_REGS 26
#define NUM_VFP_REGS 48 /* 16 status + 32 data */
typedef struct
{
  int reg_num;
  int mode;
} gdb_regmap;

static const char hexchars[16] = "0123456789abcdef";

/* These must be kept in sync with gdb/signals.h within the gnu source
   tree.  */
enum target_signal
{
  GDB_SIGNAL_INT	= 2,
  GDB_SIGNAL_ILL	= 4,
  GDB_SIGNAL_TRAP	= 5,
  GDB_SIGNAL_ABRT	= 6,
  GDB_SIGNAL_SEGV	= 11,
  GDB_SIGNAL_KILL	= 9
};

typedef struct
{
  RDI_PointHandle handle;
  uint32_t addr;
  int type;
} gdb_point;

struct rdi_stub_state
{
#ifdef __WIN32
  HANDLE dsohandle;
#else
  void * dsohandle;
#endif

  RDI_ProcVec *rdivec;
  RDI_AgentHandle rdi_agent;
  struct RDI_HostosInterface rdi_hostif;
  unsigned char fileio_data;

  unsigned rdi_modulecount;
  RDI_ModuleDesc *rdi_moduledesc;
  const RDI_ProcVec **rdi_procvec;
  RDI_ModuleHandle rdi_armmodule;
  const RDI_ProcVec *rdi_armprocvec;
  int rdi_arm_moduleid;
  int rdi_vfp_moduleid;
  uint32_t rdi_blocksize;
  uint32_t rdi_pointcaps;

  gdb_point *rdi_points;
  int rdi_point_count;
  int rdi_points_allocated;

  SOCKET gdb_listensock;
  SOCKET gdb_socket;
  struct sockaddr_in gdb_listenaddr;

  char *gdb_response;
  int gdb_response_size;
  int gdb_response_buflen;

  int gdb_vfp_regnum;
  gdb_regmap vfp_regmap[NUM_VFP_REGS];

  int verbose;
  unsigned silent:1;
  unsigned connect:1;
  unsigned running:1;
};

struct rdi_stub_state g;

static void cleanup (void);
static void do_stuff (void);
static void add_response (const unsigned char *, int);
static void send_response (void);

void __attribute__((noreturn))
die (char * fmt, ...)
{
  va_list va;

  fprintf (stderr, "Error:");
  va_start (va, fmt);
  vfprintf (stderr, fmt, va);
  va_end (va);
  fprintf (stderr, "\n");
  cleanup ();
  exit (1);
}


/* Display the error message.  */
/* TODO: Not all host errors are fatal.  Need to audit all callers.  */
static void __attribute__((noreturn))
die_errno (int err, char * msg)
{
  char buf[256];
  int len;

  if (err >= RDIError_TargetErrorBase && err <=RDIError_TargetErrorTop)
    {
      len = g.rdivec->errmess (g.rdi_agent, buf, 255, err);
      buf[len] = 0;
      die ("%s:%s", msg, buf);
    }
  else
    {
      die ("%s:RDI error %d\n", msg, err);
    }
}


/* printf clone that only generates output when verbose.  */
void
verbose (char * fmt, ...)
{
  va_list va;

  if (!g.verbose)
    return;
  va_start (va, fmt);
  vfprintf (stderr, fmt, va);
  va_end (va);
}


/* printf clone that only generates output when not in silent mode.  */
void
say (char * fmt, ...)
{
  va_list va;

  if (g.silent)
    return;
  va_start (va, fmt);
  vfprintf (stderr, fmt, va);
  va_end (va);
}


/* Host interface support routines.  */

static void
hif_dbgprint (RDI_Hif_DbgArg * arg, const char * format, va_list ap)
{
  if (!g.silent)
    vprintf (format, ap);
}

static void
hif_dbgpause (RDI_Hif_DbgArg * arg)
{
  if (!g.silent)
    fgetc(stdin);
}

/* Print a character to stdout on the gdb host debugger terminal.
   Must only be called when gdb is waiting for the response to a
   's' or 'c' command.  */
static void
gdb_print_char (int c)
{
  g.fileio_data = c;
  add_response ("Fwrite,1,ffffffff,1", 19);
  send_response ();
  do_stuff ();
}

static void
hif_writec (RDI_Hif_HostosArg * arg, int c)
{
  if (g.running)
    gdb_print_char (c);
  else
    fputc (c, stdout);
}

static int
hif_readc (RDI_Hif_HostosArg * arg)
{
  return fgetc (stdin);
}

static int
hif_write (RDI_Hif_HostosArg * arg, const char * buffer, int len)
{
  if (len > RDI_HIF_BLOCKSIZE)
    len = RDI_HIF_BLOCKSIZE;
  if (g.running)
    {
      int i;

      for (i = 0; i < len; i++)
	gdb_print_char (buffer[i]);
      return len;
    }
  else
    return fwrite (buffer, 1, len, stdout);
}

static char *
hif_gets (RDI_Hif_HostosArg * arg, char * buffer, int len)
{
  return fgets(buffer, len, stdin);
}

static void
hif_reset (RDI_Hif_ResetArg * arg)
{
  /* TODO: Should we do anything here?  Print a blank line maybe?  */
}

static RDI_Hif_UserMessageReturn
hif_message (RDI_Hif_UserMessageArg * arg, const char * msg, uint32 type)
{
  if (type == RDI_MB_STATUS)
    return RDI_MBANSWER_IGNORE;

  fputs (msg, stdout);
  /* TODO: We need a machanism for asking the user.
     Pick an arbutary response instead.  */
  switch (type)
    {
    case RDI_MB_OK:
    case RDI_MB_OKCANCEL:
      fputs ("\nAssuming OK\n", stdout);
      return RDI_MBANSWER_OK;

    case RDI_MB_YESNO:
    case RDI_MB_YESNOCANCEL:
      fputs ("\nAssuming Yes\n", stdout);
      return RDI_MBANSWER_YES;

    case RDI_MB_RETRYCANCEL:
      fputs ("\nNot retrying\n", stdout);
      return RDI_MBANSWER_CANCEL;

    case RDI_MB_ABORTRETRYIGNORE:
      fputs ("\nIgnoring\n", stdout);
      return RDI_MBANSWER_IGNORE;

    default:
      /* SHould never happen.  */
      return 0;
    }
}

/* Initialize the RDI host interface.  */

static void
init_hostif (struct RDI_HostosInterface * hif)
{
  hif->dbgprint = hif_dbgprint;
  hif->dbgpause = hif_dbgpause;
  hif->dbgarg = NULL;

  hif->writec = hif_writec;
  hif->readc = hif_readc;
  hif->write = hif_write;
  hif->gets = hif_gets;
  hif->hostosarg = NULL;

  hif->reset = hif_reset;
  hif->resetarg = NULL;

  hif->message = hif_message;
  hif->messagearg = NULL;
}


/* Setup the toolconf structure, reading settings from a file.  */

static toolconf
init_toolconf (const char * file)
{
  toolconf tc;

  tc = ToolConf_New (-1);
  if (file)
    tc = ToolConf_ReadFromFile (tc, file);

  return tc;
}


/* Open the shared object and obtain the QuerySORDI routine.  */
static QuerySORDI_Proc *
init_dso (const char * name)
{
  QuerySORDI_Proc *queryfn;

#ifdef __WIN32
  g.dsohandle = LoadLibrary(name);
#else
  g.dsohandle = dlopen (name, RTLD_LAZY);
#endif  

  if (!g.dsohandle)
    die ("library could not be loaded");

#ifdef __WIN32
  queryfn = (QuerySORDI_Proc *)GetProcAddress(g.dsohandle, "QuerySORDI"); 
#else
  queryfn = (QuerySORDI_Proc *)dlsym (g.dsohandle, "QuerySORDI");
#endif

  if (!queryfn)
    die ("QuerySORDI not found");

  return queryfn;
}


/* Open the shared object and get the RDI_ProcVec.  */

static void
init_sordi (const char * name, toolconf tc)
{
  QuerySORDI_Proc *queryfn;
  const SORDI_ProcVec *vec;
  int ver;
  uint32_t capability;

  queryfn = init_dso (name);

  vec = queryfn ();
  if (!vec)
    die ("QuerySORDI failed");

  if (!vec->isvalid || !vec->isvalid ())
    die ("Not a valid SORDI object");

  if (vec->setversion)
    {
      verbose ("Setting version to %d\n", RDI_VERSION);
      vec->setversion (RDI_VERSION);
    }

  if (!vec->getversion)
    die ("SORDI_GetVersion does not exist");
  ver = vec->getversion ();
  verbose ("RDI version %d.%d.%d\n", ver / 100, (ver / 10) % 10, ver % 10);
  if (ver < 150)
    die ("RDI version 1.5.0 or later required");

  /* TODO: We should probably call SetVersion.  */

  if (!vec->description)
    die ("SORDI_Get_DLL_Description does not exist");
  verbose ("RDI module description:%s\n", vec->description ());

  capability = RDIControllerCapability_SDMModules;
  vec->info(tc, SORDIInfo_ControllerCapabilities, &capability, NULL);

  if (!vec->getrdi ())
    die ("SORDI_GetRDIProcVec does not exist");

  g.rdivec = vec->getrdi ();
  if (!g.rdivec)
    die ("SORDI_GetRDIProcVec failed");

  if (vec->init)
    {
      const char *errstring;

      if (!vec->init (&errstring, tc))
	die ("SORDI_Initialise failed: %s", errstring);
    }

  /* TODO: SORDI_Info for SDM.  */
  verbose ("SORDI initialised\n");
}


/* Unload the SORDI shared object.  */

static void
cleanup_dso ()
{
  if (!g.dsohandle)
    return;

#ifdef __WIN32
  FreeLibrary(g.dsohandle);
#else
  dlclose(g.dsohandle);
#endif

  g.dsohandle = NULL;
}


/* Perform module specific initialization.  */

static void
rdi_process_module (int n)
{
  unsigned nregs;
  RDIRegister *regs;
  int rc;
  int i;
  int gdb_reg;
  const char *name;
  int next_gdb_regnum;

  if (strcmp (g.rdi_moduledesc[n].type, "ARM") == 0)
    {
      g.rdi_armprocvec = g.rdi_procvec[n];
      g.rdi_armmodule = g.rdi_moduledesc[n].handle;
      g.rdi_arm_moduleid = n;
    }
  rc = g.rdi_procvec[n]->info (g.rdi_moduledesc[n].handle,
      RDIInfo_DescribeRegisterMap, (ARMword *) &nregs, (ARMword *) &regs);
  if (rc != RDIError_NoError)
    return;

  verbose ("Register Name/Bank (num, mode) size attrib type\n");
  for (i = 0; i < nregs; i++)
    {
      verbose ("%d: %s/%s (%d, %d) %d %x %s\n",
	       i, regs[i].name, regs[i].bankname,
	       regs[i].reg_num, regs[i].mode, regs[i].width,
	       regs[i].attrib, regs[i].displaytype);
    }

  next_gdb_regnum = NUM_GDB_G_REGS;
  /* TODO: Is there a better way to recognise VFP units?  */
  if (strncmp (g.rdi_moduledesc[n].name, "VFP", 3) == 0)
    {
      g.rdi_vfp_moduleid = n;
      g.gdb_vfp_regnum = next_gdb_regnum;
      next_gdb_regnum += NUM_VFP_REGS;
      for (gdb_reg = 0; gdb_reg < NUM_VFP_REGS; gdb_reg++)
	g.vfp_regmap[gdb_reg].reg_num = -1;

      verbose ("Allocated VFP gdb regnum %d\n", g.gdb_vfp_regnum);
      for (i = 0; i < nregs; i++)
	{
	  name = regs[i].name;
	  if (name[0] == 'S' && isdigit(name[1]))
	    {
	      if (name[2])
		gdb_reg = (name[1] - '0') * 10 + (name[2] - '0');
	      else
		gdb_reg = name[1] - '0';
	      gdb_reg += 16;
	    }
	  else if (strcmp(name, "FPSID") == 0)
	    gdb_reg = 0;
	  else if (strcmp (name, "FPSCR") == 0)
	    gdb_reg = 1;
	  else if (strcmp (name, "FPEXC") == 0)
	    gdb_reg = 8;
	  else
	    gdb_reg = -1;

	  if (gdb_reg != -1)
	    {
	      g.vfp_regmap[gdb_reg].reg_num = regs[i].reg_num;
	      g.vfp_regmap[gdb_reg].mode = regs[i].mode;
	    }
	}
    }
}


/* Open RDI modules.  */

static void 
init_rdi_modules (toolconf tc)
{
  int rc;
  int i;
  unsigned nmodules;
  unsigned type;

  /* Get the number of modules.  We store this locally, and only update the
     global state when we have actually opened the module.  This ensures
     that we do the right thing if opening one of the modules fails.  */
  rc = g.rdivec->info (g.rdi_agent, RDIInfo_Modules, (ARMword *)&nmodules,
		       NULL);
  if (rc != RDIError_NoError)
    die_errno (rc, "Unable to determine number of processors");
  verbose ("Found %d module(s)\n", nmodules);

  /* Retrieve information about the modules.  */
  g.rdi_moduledesc = (RDI_ModuleDesc *)
    malloc(nmodules * sizeof(RDI_ModuleDesc));
  g.rdi_procvec = (const RDI_ProcVec **)
    malloc (nmodules * sizeof (RDI_ProcVec *));
  if (!g.rdi_moduledesc || !g.rdi_procvec)
    die ("out of memory");

  rc = g.rdivec->info (g.rdi_agent, RDIInfo_Modules, (ARMword *)&nmodules,
		       (ARMword *)g.rdi_moduledesc);
  if (rc != RDIError_NoError)
    die_errno (rc, "RDIInfo_Modules failed");

  type = (1 << RDIOpen_CommsReset)
	 | (RDISex_DontCare << RDIOpen_ByteSexShift);
  /* Open the modules.  */
  for (i = 0; i < nmodules; i++)
    {
      if (g.rdi_moduledesc[i].rdi)
	g.rdi_procvec[i] = g.rdi_moduledesc[i].rdi;
      else
	g.rdi_procvec[i] = g.rdivec;
      verbose ("Module %.8s:%.8s\n", g.rdi_moduledesc[i].type,
	       g.rdi_moduledesc[i].name);
      rc = g.rdi_procvec[i]->open (g.rdi_moduledesc[i].handle, type, tc,
				&g.rdi_hostif, NULL);
      switch (rc)
	{
	case RDIError_NoError:
	case RDIError_LittleEndian:
	case RDIError_BigEndian:
	  /* TODO: record target endianness.  */
	  verbose ("Opened RDI module\n");
	  break;

	default:
	  die_errno (rc, "Unable to open RDI module");
	}
      /* Module has been initialised.  Update global state to reflect this.  */
      g.rdi_modulecount = i + 1;

      rdi_process_module (i);
    }

  if (!g.rdi_armmodule)
    die ("No ARM processor RDI module found");
}


/* Check for RDI features.  */
static void
init_rdi_caps (void)
{
  int i;
  int first;
  static const char* const names [13] = {
      "Comparison",
      "Range",
      "ByteRead",
      "HalfRead",
      "WordRead",
      "ByteWrite",
      "HalfWrite",
      "WordWrite",
      "Mask",
      "ThreadBreak",
      "ThreadWatch",
      "CondBreak",
      "Status"};
  g.rdi_pointcaps = 0;
  g.rdi_armprocvec->info (g.rdi_armmodule, RDIInfo_Points,
			  (ARMword *) &g.rdi_pointcaps, NULL);
  verbose ("RDIInfoPoints: %x (", g.rdi_pointcaps);
  first = 1;
  for (i = 0; i < 13; i++)
    {
      if (g.rdi_pointcaps & (1 << i))
	{
	  verbose ("%s%s", first ? "" : ", ", names[i]);
	  first = 0;
	}
    }
  verbose (")\n");
}

/* Open the target interface.  */

static void
init_rdi (const char * dsoname, const char *configfile)
{
  int rc;
  unsigned type;
  toolconf tc;

  tc = init_toolconf (configfile);
  init_sordi (dsoname, tc);
  init_hostif (&g.rdi_hostif);

  type = (1 << RDIOpen_CommsReset)
	 | (RDISex_DontCare << RDIOpen_ByteSexShift);
  rc = g.rdivec->openagent (&g.rdi_agent, type, tc, &g.rdi_hostif, NULL);
  switch (rc)
    {
    case RDIError_NoError:
    case RDIError_LittleEndian:
    case RDIError_BigEndian:
      /* TODO: record target endianness.  */
      verbose ("Opened RDI agent\n");
      break;

    default:
      die_errno (rc, "Unable to open RDI agent");
    }

  init_rdi_modules (tc);

  init_rdi_caps ();

  rc = g.rdi_armprocvec->info (g.rdi_armmodule, RDIInfo_GetLoadSize,
      (ARMword *)&g.rdi_blocksize, NULL);
  switch (rc)
    {
    case RDIError_NoError:
      break;

    case RDIError_UnimplementedMessage:
      /* Pick something sensible.  */
      g.rdi_blocksize = 1024;
      break;

    default:
      die_errno (rc, "RDIInfo_GetLoadSize failed");
      break;
    }

  {
    /* TDOD: Document which exceptions we are catching.  */
    uint32_t flags = 0x3a;
    rc = g.rdi_armprocvec->info (g.rdi_armmodule, RDIVector_Catch,
				 (ARMword *) &flags, NULL);
  }
  verbose ("RDI initialization complete\n");
}

/* Initialise the socket subsystem.  */

static void
init_sockets (void)
{
#ifdef HAVE_WINSOCK2_H
  WSADATA wsa_data;
  if (WSAStartup (MAKEWORD(1, 0), &wsa_data) != 0)
    die ("WSAStartup failed");
#else
  /* Do nothing. */
#endif
}

/* Finalize the socket subsystem.  */

static void
cleanup_sockets (void)
{
#ifdef HAVE_WINSOCK2_H
  WSACleanup ();
#else
  /* Do nothing. */
#endif
}

/* Create a socket, and listen on the specified address.  */

static void
init_listen_sock (void)
{
  SOCKET s;
  int port;

  /* Do nothing if we're initiating the connection.  */
  if (g.connect)
    return;

  s = socket (PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (!SOCKET_OK (s))
    die ("Unable to create socket");

  g.gdb_listensock = s;

  port = ntohs (g.gdb_listenaddr.sin_port);
  if (bind (s, (struct sockaddr *) &g.gdb_listenaddr,
	    sizeof (g.gdb_listenaddr)))
    die ("Unable to bind to port %d", port);

  if (listen (s, 1))
    die ("Listen on socket failed");

  say ("Listening for connections on %s:%d\n",
       inet_ntoa (g.gdb_listenaddr.sin_addr), port);
}

/* Close gdb listen socket.  */
static void
cleanup_listensock (void)
{
  if (!g.gdb_listensock)
    return;

  closesocket (g.gdb_listensock);
  g.gdb_listensock = 0;

  verbose ("Socket closed\n");
}

/* Close the active gdb connection.  */
static void
cleanup_connection (void)
{
  char c;

  if (!g.gdb_socket)
    return;

  verbose ("Closing gdb connection\n");

  shutdown (g.gdb_socket, SD_SEND);
  while (read (g.gdb_socket, &c, 1) > 0)
    /* nop */ ;
  closesocket (g.gdb_socket);

  g.gdb_socket = 0;
  verbose ("Connection closed\n");
}

static void
cleanup_gdb (void)
{
  cleanup_connection ();
  cleanup_listensock ();
}

/* Setup the gdb remote interface.  */
void
init_gdb (void)
{
  init_listen_sock ();
}

/* Close RDI modules. */

void
cleanup_rdi_modules (void)
{
  int i;
  int rc;

  while (g.rdi_modulecount > 0)
    {
      i = --g.rdi_modulecount;

      rc = g.rdi_procvec[i]->close (g.rdi_moduledesc[i].handle);
      if (rc != RDIError_NoError)
	die ("Unable to close module");
    }
}


/* Close RDI connections.  */

void
cleanup_rdi (void)
{
  int rc;

  cleanup_rdi_modules ();

  if (g.rdi_agent)
    {
      rc = g.rdivec->closeagent (g.rdi_agent);
      g.rdi_agent = NULL;
      if (rc != RDIError_NoError)
	die ("Unable to close RDI agent");
    }

  cleanup_dso ();
}


/* Shut everything down.  May be called recursively if an error occurs.  */

static void
cleanup (void)
{
  cleanup_rdi ();
  cleanup_gdb ();
  cleanup_sockets ();
}

/* Initiate a gdb connection.  */
static void
initiate_connection (void)
{
  SOCKET s;
  int tmp;

  s = socket (PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (!SOCKET_OK (s))
    die ("Unable to create socket");

  g.gdb_socket = s;

  if (connect (s, (struct sockaddr *) &g.gdb_listenaddr,
	       sizeof (g.gdb_listenaddr)))
    die ("Unable to connect\n");

  /* Tell TCP not to delay small packets.  This greatly speeds up
     interactive response. */
  tmp = 1;
  setsockopt (s, IPPROTO_TCP, TCP_NODELAY,
	      (char *) &tmp, sizeof (tmp));


  say ("Connected\n");
}

/* Wait for a gdb remote connection.  */
static void
get_connection (void)
{
  struct sockaddr_in sa;
  SOCKET s;
  int tmp;
  socklen_t len;
  
  len = sizeof(sa);
  s = accept (g.gdb_listensock, (struct sockaddr *) &sa, &len);
  if (!SOCKET_OK (s))
    die ("Incoming connection failed");

  /* Tell TCP not to delay small packets.  This greatly speeds up
     interactive response. */
  tmp = 1;
  setsockopt (s, IPPROTO_TCP, TCP_NODELAY,
	      (char *) &tmp, sizeof (tmp));

  say ("Conection accepted from %s:%d\n",
       inet_ntoa (sa.sin_addr), sa.sin_port);
  g.gdb_socket = s;

  closesocket (g.gdb_listensock);
  g.gdb_listensock = 0;
}

/* Allocate space for n additional bytes in the response buffer.  */
static void
allocate_response_buff (int n)
{
  if (g.gdb_response_buflen >= g.gdb_response_size + n)
    return;

  if (g.gdb_response_buflen == 0)
    g.gdb_response_buflen = 1024;
  while (g.gdb_response_buflen < g.gdb_response_size + n)
    g.gdb_response_buflen *= 2;

  g.gdb_response = realloc (g.gdb_response, g.gdb_response_buflen);
  if (!g.gdb_response)
    die ("Out of memory");
}

/* Calculate the checksum for a block of data.  */
static unsigned char
gdb_checksum(const unsigned char * data, int len)
{
  unsigned char csum;

  csum = 0;
  while (len--)
    csum += *(data++);

  return csum;
}

/* Initialise a new response packet.  */
static void
clear_response (void)
{
  if (g.gdb_response_size)
    die ("internal error: multiple response packets");

  allocate_response_buff (1);
  g.gdb_response_size = 1;
  g.gdb_response[0] = '$';
}

/* Calculate checksum and send response packet.  */
static void
send_response (void)
{
  unsigned char csum;
  int i;

  csum = gdb_checksum (&g.gdb_response[1], g.gdb_response_size - 1);
  allocate_response_buff (3);
  g.gdb_response[g.gdb_response_size] = '#';
  g.gdb_response[g.gdb_response_size + 1] = hexchars[csum >> 4];
  g.gdb_response[g.gdb_response_size + 2] = hexchars[csum & 0xf];
  i = send (g.gdb_socket, g.gdb_response, g.gdb_response_size + 3, 0);
  /* TODO: Is this always fatal?  */
  if (i < g.gdb_response_size)
    die ("transmit error");
  verbose ("Sent response: '%.*s'\n", g.gdb_response_size - 1,
	   &g.gdb_response[1]);
}

/* Resend a response packet.  */
static void
resend_response (void)
{
  int i;

  i = send (g.gdb_socket, g.gdb_response, g.gdb_response_size + 3, 0);
  /* TODO: Is this always fatal?  */
  if (i < g.gdb_response_size)
    die ("transmit error");
  verbose ("Resent response\n");
}

/* Add a data to the current response.  */
static void
add_response (const unsigned char * data, int len)
{
  allocate_response_buff (len);
  memcpy (&g.gdb_response[g.gdb_response_size], data, len);
  g.gdb_response_size += len;
}

/* Add ascii hex repesentation of a 32-bit value to the output.  */
static void
add_response_32 (uint32_t val)
{
  unsigned char buff[8];
  int i;

  /* TODO: big-endian targets.  */
  for (i = 0; i < 4; i++)
    {
      buff[i * 2 + 1] = hexchars[val & 0xf];
      val >>= 4;
      buff[i * 2] = hexchars[val & 0xf];
      val >>= 4;
    }
  add_response (buff, 8);
}

static void
write_hex_byte (int c)
{
  char s[2];

  s[0] = hexchars[c >> 4];
  s[1] = hexchars[c & 0xf];
  add_response (s, 2);
}

/* Convert a single hex character into an integer.  */
static int
from_shex (char c)
{
  if (c >= '0' && c <= '9')
    return c - '0';
  else if (c >= 'A' && c <= 'Z')
    return c + 10 - 'A';
  else if (c >= 'a' && c <= 'z')
    return c + 10 - 'a';
  else
    die ("internal error: invalid hex");
}

/* Read an ascii hex number.  Returns a pointer to the following
   character.  */
static unsigned char *
read_hex (const unsigned char * s, uint32_t * p)
{
  uint32_t val;

  val = 0;
  while (isxdigit (*s))
    {
      val = (val << 4) | from_shex(*s);
      s++;
    }
  *p = val;
  return (unsigned char *) s;
}
   
/* Convert an 8-byte ascii hex string into an integer.  Returns
   zero on success, nonzero on failure.  Target byte order.  */
static int
read_hex32 (const unsigned char * data, uint32_t * p)
{
  int i;
  uint32_t val;
  uint32_t n;

  val = 0;
  /* TODO: big-endian targets.  */
  for (i = 0; i < 4; i++)
    {
      if (!isxdigit (data[i * 2])
	  || !isxdigit (data[i * 2 + 1]))
	return 1;
      n = from_shex (data[i * 2 + 1])
	| (from_shex (data[i * 2]) << 4);
      val |= n << (i * 8);
    }
  *p = val;
  return 0;
}

struct gdb_callback
{
  char c;
  void (*fn) (const unsigned char *, int);
};

static void gdb_cmd_status (const unsigned char *, int);
static void gdb_cmd_read_regs (const unsigned char *, int);
static void gdb_cmd_write_regs (const unsigned char *, int);
static void gdb_cmd_read_mem (const unsigned char *, int);
static void gdb_cmd_write_mem (const unsigned char *, int);
static void gdb_cmd_step (const unsigned char *, int);
static void gdb_cmd_continue (const unsigned char *, int);
static void gdb_cmd_addpoint (const unsigned char *, int);
static void gdb_cmd_rmpoint (const unsigned char *, int);
static void gdb_cmd_read_one_reg (const unsigned char *, int);
static void gdb_cmd_write_one_reg (const unsigned char *, int);
static void gdb_cmd_query (const unsigned char *, int);

static const struct gdb_callback
gdb_callbacks[] =
{
  {'?', gdb_cmd_status},
  {'g', gdb_cmd_read_regs},
  {'G', gdb_cmd_write_regs},
  {'m', gdb_cmd_read_mem},
  {'M', gdb_cmd_write_mem},
  {'s', gdb_cmd_step},
  {'c', gdb_cmd_continue},
  {'Z', gdb_cmd_addpoint},
  {'z', gdb_cmd_rmpoint},
  {'p', gdb_cmd_read_one_reg},
  {'P', gdb_cmd_write_one_reg},
  {'q', gdb_cmd_query},
  {0, NULL}
};


/* Respond to a 'q' command.  */

static void
gdb_cmd_query (const unsigned char * packet, int len)
{
  char buf[8]; 
  int buflen;
  uint32_t addr;
  uint32_t addrlen;

  if (len > 30
      && strncmp (packet, "Part:availableRegisters:read::", 30) == 0)
    {
      if (g.gdb_vfp_regnum)
	buflen = sprintf (buf, "vfp:%x", g.gdb_vfp_regnum);
      else
	buflen = 0;
      packet += 30;
      packet = read_hex (packet, &addr);
      if (*(packet++) != ',')
	goto bad;
      packet = read_hex (packet, &addrlen);
      if (*packet != '#')
	goto bad;
      
      if (addr > buflen)
	goto bad;
      if (addr + addrlen > buflen)
	addrlen = buflen - addr;

      if (addrlen == 0)
	{
	  add_response ("OK", 2);
	  return;
	}
      while (addrlen > 0)
	{
	  write_hex_byte (buf[addr]);
	  addr++;
	  addrlen--;
	}
    }
  return;
bad:
  add_response ("E22", 3);
}


/* Respond to a '?' command.  */
static void
gdb_cmd_status (const unsigned char * packet, int len)
{
  /* TODO: Report something sensible.  */
  add_response ("S00", 3);
}


/* Set the value of a register.  Used by transfer_reg.  */
/* This currently only handles 32-bit registers.  */

static int
set_reg (int moduleid, int reg, int mode,
	 const unsigned char * packet)
{
  int rc;
  uint32_t val;

  if (read_hex32 (packet, &val))
    return 1;
  if (packet[8] != '#')
    return 1;

  rc = g.rdi_procvec[moduleid]->CPUwrite (g.rdi_moduledesc[moduleid].handle,
      mode, 1u << reg, (ARMword *) &val);
  if (rc != RDIError_NoError)
    return 1;
  return 0;
}


/* Get a register value.  Used by transfer_reg.  */
/* This currently only handles 32-bit registers.  */

static int
get_reg (int moduleid, int reg, int mode)
{
  int rc;
  uint32_t val;
  int i;

  rc = g.rdi_procvec[moduleid]->CPUread (g.rdi_moduledesc[moduleid].handle,
      mode, 1u << reg, (ARMword *) &val);
  if (rc != RDIError_NoError)
    return 1;

  /* TODO: big-endian targets.  */
  for (i = 0; i < 4; i++)
    {
      write_hex_byte (val & 0xff);
      val >>= 8;
    }
  return 0;
}


/* Read or write a register.  Helper function for gdb_cmd_{read,write}_one_reg.
   Returns zero on success, 1 on failure and -1 if the register is not
   supported.  If packet is NULL, add the register contents to the gdb
   response packet.  Otherwise packet points at the register contents in
   ascii hex form.  */
static int
transfer_reg (uint32_t reg, const unsigned char * packet)
{
  uint32_t val;
  uint32_t rdi_reg;
  int i;
  int rc;

  if ((reg >= 0 && reg <= 15) || reg == 25)
    {
      /* Core registers.  */
      switch (reg)
	{
	case 15:
	  rdi_reg = 16;
	  break;

	case 25:
	  rdi_reg = 17;
	  break;

	default:
	  rdi_reg = reg;
	  break;
	}
      if (packet)
	return set_reg (g.rdi_arm_moduleid, rdi_reg, 0xff, packet);
      else
	return get_reg (g.rdi_arm_moduleid, rdi_reg, 0xff);
    }
  else if (reg == 24)
    {
      /* FPA scr */
      /* TODO: implement fpa.  */
      if (!packet)
	add_response ("x", 1);
      return -1;
    }
  else if (reg >= 16 && reg <= 23)
    {
      /* FPA register.  */
      /* TODO: implement fpa.  */
      if (!packet)
	add_response ("x", 1);
      return -1;
    }
  else if (g.gdb_vfp_regnum && reg >= g.gdb_vfp_regnum
	   && reg <= g.gdb_vfp_regnum + NUM_VFP_REGS)
    {
      gdb_regmap *r;

      r = &g.vfp_regmap[reg - g.gdb_vfp_regnum];
      /* VFP register.  */
      if (r->reg_num == -1)
	{
	  if (!packet)
	    add_response ("x", 1);
	  return -1;
	}
      if (packet)
	return set_reg (g.rdi_vfp_moduleid, r->reg_num, r->mode, packet);
      else
	return get_reg (g.rdi_vfp_moduleid, r->reg_num, r->mode);
    }
  else
    {
      if (!packet)
	add_response ("x", 1);
      return -1;
    }
}

/* Respond to a 'p' command.  */
static void
gdb_cmd_read_one_reg (const unsigned char * packet, int len)
{
  uint32_t reg;

  packet = read_hex (packet, &reg);
  if (*(packet++) != '#')
    goto bad;

  if (transfer_reg (reg, NULL) > 0)
    goto bad;

  return;
bad:
  verbose ("Bad 'p' packet\n");
  add_response ("E22", 3);
}

/* Respond to a 'P' command.  */
static void
gdb_cmd_write_one_reg (const unsigned char * packet, int len)
{
  uint32_t reg;

  packet = read_hex (packet, &reg);
  if (*(packet++) != '=')
    goto bad;

  if (transfer_reg (reg, packet))
    goto bad;

  add_response ("OK", 2);
  return;
bad:
  verbose ("Bad 'P' packet\n");
  add_response ("E22", 3);
}

/* Respond to a 'g' command.  */
static void
gdb_cmd_read_regs (const unsigned char * packet, int len)
{
  uint32_t int_regs[17];
  int rc;
  int i;

  if (len > 0)
    verbose ("Unexpected junk after g command");
  
  /* The gdb remote protocol wants r0-15, f0-f7, fp and cpsr.  */
  rc = g.rdi_armprocvec->CPUread (g.rdi_armmodule, 0xff, 0x0037fff,
				  (ARMword *)int_regs);
  /* TODO: Fail gracefully.  */
  if (rc != RDIError_NoError)
    die_errno (rc, "CPUread failed");

  /* Output arm registers.  */
  for (i = 0; i < 16; i++)
    add_response_32 (int_regs[i]);
  /* Output zeros for rf0-f7.  */
  for (i = 0; i < 8; i++)
    add_response ("000000000000000000000000", 24);
  /* Output zeroes for fps.  */
  add_response ("00000000", 8);
  /* Output cpsr.  */
  add_response_32 (int_regs[16]);
}

/* Respond to a 'G' command.  */
static void
gdb_cmd_write_regs (const unsigned char * packet, int len)
{
  uint32_t int_regs[17];
  int i;
  int rc;

  if (len != 336)
    {
      verbose ("Bad 'G' packet length\n");
      add_response ("E22", 3);
      return;
    }

  /* Core integer regs.  */
  for (i = 0; i < 16; i++)
    {
      if (read_hex32 (packet, &int_regs[i]))
	{
	  verbose ("Bad hex in 'G' packet\n");
	  add_response ("E22", 3);
	  return;
	}
      packet += 8;
    }
  /* Ignore the fpa regs.  */
  packet += 24 * 8 + 8;
  /* cpsr.  */
  if (read_hex32 (packet, &int_regs[16]))
    {
      verbose ("Bad hex in 'G' packet\n");
      add_response ("E22", 3);
      return;
    }
  rc = g.rdi_armprocvec->CPUwrite (g.rdi_armmodule, 0xff, 0x0037fff,
				   (ARMword *)int_regs);
  if (rc != RDIError_NoError)
    {
      /* TODO: Check for fatal target errors.  */
      verbose ("CPUwrite failed\n");
      add_response ("E16", 3);
      return;
    }
  add_response ("OK", 2);
}

/* Respond to a 'm' command.  */
static void
gdb_cmd_read_mem (const unsigned char * packet, int len)
{
  uint32_t addr;
  uint32_t addrlen;
  unsigned size;
  unsigned char * buff;
  unsigned char * p;
  int rc;

  packet = read_hex (packet, &addr);
  if (*(packet++) != ',')
    {
      verbose ("Expected ',' in 'm' command\n");
      add_response ("E22", 3);
      return;
    }
  packet = read_hex (packet, &addrlen);
  if (*packet != '#')
    {
      verbose ("Unexpected junk at end of 'm' command\n");
      add_response ("E22", 3);
      return;
    }

  size = (addrlen > g.rdi_blocksize) ? g.rdi_blocksize : addrlen;
  buff = malloc (size);
  while (addrlen > 0)
    {
      p = buff;
      size = (addrlen > g.rdi_blocksize) ? g.rdi_blocksize : addrlen;
      /* TODO: We don't know if this is data or code.  */
      rc = g.rdi_armprocvec->read (g.rdi_armmodule, addr, p, &size,
				   RDIAccess_Data);
      switch (rc)
	{
	case RDIError_NoError:
	case RDIError_BigEndian:
	case RDIError_LittleEndian:
	  /* TODO: Handle endianness change.  */
	  break;

	case RDIError_DataAbort:
	  verbose ("Data abort during 'm' command\n");
	  add_response ("E14", 3);
	  free (buff);
	  return;

	case RDIError_UnimplementedSize:
	  /* Should never happen.  */
	  verbose ("RDI_ReadProc: Unimplemented size\n");
	  /* Fall through.  */

	default:
	  die_errno (rc, "RDI_ReadProc failed");
	}
      if (size <= 0)
	die ("RDI_ReadProc returned no bytes");

      addrlen -= size;
      while (size--)
	{
	  write_hex_byte (*(p++));
	  addr++;
	}
    }
  free(buff);
}

/* Respond to a 'M' command.  */
static void
gdb_cmd_write_mem (const unsigned char * packet, int len)
{
  uint32_t addr;
  uint32_t addrlen;
  unsigned size;
  unsigned n;
  char * buff;
  char * p;
  int rc;

  packet = read_hex (packet, &addr);
  if (!(packet && *(packet++) == ','))
    {
      verbose ("Expected ',' in 'M' command\n");
      add_response ("E22", 3);
      return;
    }
  packet = read_hex (packet, &addrlen);
  if (!(packet && *(packet++) == ':'))
    {
      verbose ("Expected ':' in 'M' command\n");
      add_response ("E22", 3);
      return;
    }

  size = (addrlen > g.rdi_blocksize) ? g.rdi_blocksize : addrlen;
  buff = malloc (size);
  while (addrlen > 0)
    {
      size = (addrlen > g.rdi_blocksize) ? g.rdi_blocksize : addrlen;
      p = buff;
      for ( n = 0; n < size; n++)
	{
	  if (!isxdigit (packet[0])
	      || !isxdigit (packet[1]))
	    {
	      verbose ("Expected hex digit in 'M' command\n");
	      add_response ("E22", 3);
	      free (buff);
	      return;
	    }
	  *(p++) = (from_shex (packet[0]) << 4) | from_shex (packet[1]);
	  packet += 2;
	}
      p = buff;
      while (size > 0)
	{
	  n = size;
	  /* TODO: We don't know if this is data or code.  */
	  rc = g.rdi_armprocvec->write (g.rdi_armmodule, p, addr, &n,
					RDIAccess_Code);
	  switch (rc)
	    {
	    case RDIError_NoError:
	    case RDIError_BigEndian:
	    case RDIError_LittleEndian:
	      /* TODO: Handle endianness change.  */
	      break;

	    case RDIError_DataAbort:
	    case RDIError_InsufficientPrivilege:
	      verbose ("Data abort during 'M' command\n");
	      add_response ("E14", 3);
	      break;

	    case RDIError_UnimplementedSize:
	      /* Should never happen.  */
	      verbose ("RDI_WriteProc: Unimplemented size\n");
	      /* Fall through.  */

	    default:
	      die_errno (rc, "RDI_WriteProc failed");
	    }
	  if (n <= 0)
	    die ("RDI_WriteProc wrote no bytes");
	  size -= n;
	  p += n;
	  addr += n;
	  addrlen -= n;

	}
    }
  free (buff);
  add_response ("OK", 2);
}

/* Add the PC value to the end of a "T" stop packet.  */
/* TODO: Do we also want to supply other register values?  */
static void
add_stop_info (void)
{
  add_response ("f:", 2);
  transfer_reg (15, NULL);
  add_response (";", 1);
}

/* Report that execution stopped with a signal.  */
static void
gdb_signal (unsigned char signum)
{
  char buff[4];

  verbose ("Got signal %d\n", signum);
  sprintf (buff, "T%02x", signum);
  add_response (buff, 3);

  add_stop_info ();
}

/* Watchpoint hit.  */
static void
gdb_sigwatch (RDI_PointHandle handle)
{
  char buff[100];
  static const char *point_type_names[3] = {"watch", "rwatch", "awatch"};
  const char *s;
  int shift;
  gdb_point *p;
  int i;

  verbose ("Warchpoint hit\n");
  sprintf (buff, "T%02x", GDB_SIGNAL_TRAP);
  add_response (buff, 3);

  p = g.rdi_points;
  for (i = 0; i < g.rdi_point_count; i++, p++)
    {
      if (p->handle == handle)
	break;
    }
  if (i == g.rdi_point_count || p->type < 2 || p->type > 4)
    {
      verbose ("Bad watchpoint\n");
      add_response (";", 1);
      return;
    }

  s = point_type_names[p->type - 2];
  add_response (s, strlen (s));
  add_response (":", 1);

  for (shift = 28; shift > 0; shift -= 4)
    {
      if (p->addr >> shift)
	break;
    }
  for (; shift >= 0; shift-= 4)
    add_response (&hexchars[(p->addr >> shift) & 0xf], 1);
  add_response (";", 1);

  add_stop_info ();
}

static uint32_t
rdi_getreg (int regno)
{
  int rc;
  uint32_t val;

  rc = g.rdi_armprocvec->CPUread (g.rdi_armmodule, 0xff, 1 << regno,
				  (ARMword *) &val);
  if (rc != RDIError_NoError)
    die_errno (rc, "CPUread failed");
  return val;
}
static void
rdi_setreg (int regno, uint32_t val)
{
  int rc;

  rc = g.rdi_armprocvec->CPUwrite (g.rdi_armmodule, 0xff, 1 << regno,
				   (ARMword *) &val);
  if (rc != RDIError_NoError)
    die_errno (rc, "CPUwrite failed");
}
/* gdb uses as undefined instruction for breakpoints.  It expects the target
   state to be just before the exception occurred.  The actual target state
   will be at entry to the trap handler.  Back out of the trap handler.  */
/* TODO: Remove or conditionalise this when we decide what to do with
   soft/hard breakpoints.  See gdb_cmd_addpoint.  */
static void
undo_trap_handler ()
{
  uint32_t regs[2];

  /* Get the handler lr and spsr.  The spsr read must be a separate call.  */
  regs[0] = rdi_getreg (14) - 4;
  regs[1] = rdi_getreg (18);
  /* Restore the cpsr and pc from saved values.  */
  rdi_setreg (17, regs[1]);
  rdi_setreg (16, regs[0]);
}

/* Report target status after execution.  */
static void
gdb_report_stopped (int rc, RDI_PointHandle handle, int step)
{
  switch (rc)
    {
    case RDIError_NoError:
    case RDIError_TargetStopped:
      verbose ("Stopped\n");
      if (step)
	gdb_signal (GDB_SIGNAL_TRAP);
      else
	{
	  verbose ("Execution stopped\n");
	  add_response ("W00", 3);
	}
      break;

    case RDIError_BreakpointReached:
      /* TODO: Is this sufficient?  */
      gdb_signal (GDB_SIGNAL_TRAP);
      break;
      
    case RDIError_WatchpointAccessed:
      gdb_sigwatch (handle);
      break;
      
    case RDIError_UserInterrupt:
      gdb_signal (GDB_SIGNAL_INT);
      break;

    case RDIError_UndefinedInstruction:
      /* TODO: Make this optional.  */
      undo_trap_handler ();
      gdb_signal (GDB_SIGNAL_ILL);
      break;

    case RDIError_SoftwareInterrupt:
      gdb_signal (GDB_SIGNAL_TRAP);
      break;

    case RDIError_PrefetchAbort:
    case RDIError_DataAbort:
    case RDIError_AddressException:
      gdb_signal (GDB_SIGNAL_SEGV);
      break;

    case RDIError_BranchThrough0:
    case RDIError_IRQ:
    case RDIError_FIQ:
      gdb_signal (GDB_SIGNAL_TRAP);
      break;

    case RDIError_UnimplementedMessage:
      die ("internal error: Unimplemented step");
      break;

    default:
      die_errno (rc, "Execution failed");
    }
}

/* Respond to a 's' command.  */
static void
gdb_cmd_step (const unsigned char * packet, int len)
{
  int rc;
  uint32_t addr;
  RDI_PointHandle handle;

  if (*packet != '#')
    {
      /* A start address was specified.  */
      packet = read_hex (packet, &addr);
      if (*packet != '#')
	{
	  verbose ("Unexpected junk at end of 's' packet\n");
	  add_response ("E22", 3);
	  return;
	}
      rdi_setreg (16, addr);
    }

  verbose ("Stepping target\n");
  g.running = 1;
  rc = g.rdivec->step (g.rdi_agent, &g.rdi_armmodule, FALSE, 1, &handle);
  g.running = 0;
  gdb_report_stopped (rc, handle, 1);
}

/* Respond to a 'c' command.  */
static void
gdb_cmd_continue (const unsigned char * packet, int len)
{
  int rc;
  RDI_PointHandle handle;
  uint32_t addr;

  if (*packet != '#')
    {
      /* A start address was specified.  */
      packet = read_hex (packet, &addr);
      if (*packet != '#')
	{
	  verbose ("Unexpected junk at end of 's' packet\n");
	  add_response ("E22", 3);
	  return;
	}
      rdi_setreg (16, addr);
    }

  verbose ("Running target\n");
  g.running = 1;
  rc = g.rdivec->execute (g.rdi_agent, &g.rdi_armmodule, FALSE, &handle);
  g.running = 0;
  gdb_report_stopped (rc, handle, 0);
}

/* Respond to a 'Z' command.  */
static void
gdb_cmd_addpoint (const unsigned char * packet, int len)
{
  int type;
  uint32_t addr;
  uint32_t addrlen;
  unsigned rdi_type;
  uint32_t rdi_dtype;
  int rc;
  RDI_PointHandle h;
  gdb_point *p;

  /* TODO: RDI doesn't seem to have a mechanism for specifying
     hardware of software breakpoints, We create both types with RDI_SetBreak
     then check we have a hardware breakpoint for Z1 breakpoints.  */
     
  switch (*packet)
    {
    case '0': /* soft breakpoint */
    case '1': /* hard breakpoint */
    case '2': /* write watchpoint */
    case '3': /* read watchpoint */
    case '4': /* access watchpoint */
      type = *packet - '0';
      break;

    default:
      return;
    }
  packet++;
  if (!(packet && *(packet++) == ','))
    goto bad;
  packet = read_hex (packet, &addr);
  if (!(packet && *(packet++) == ','))
    goto bad;
  packet = read_hex (packet, &addrlen);
  if (!(packet && *(packet++) == '#'))
    goto bad;

  rdi_type = RDIPoint_EQ;
  if (type < 2)
    {
      switch (addrlen)
	{
	case 2:
	  rdi_type |= RDIPoint_16Bit;
	  break;

	case 4:
	  break;

	default:
	  goto bad;
	}
      rc = g.rdi_armprocvec->setbreak (g.rdi_armmodule, addr, rdi_type,
				       0, RDI_NoHandle, &h);
    }
  else
    {
      rdi_dtype = 0;
      switch (addrlen)
	{
	case 1:
	  if (type != 2)
	    rdi_dtype |= RDIWatch_ByteRead;
	  if (type != 3)
	    rdi_dtype |= RDIWatch_ByteWrite;
	  break;

	case 2:
	  if (type != 2)
	    rdi_dtype |= RDIWatch_HalfRead;
	  if (type != 3)
	    rdi_dtype |= RDIWatch_HalfWrite;
	  break;

	case 4:
	  if (type != 2)
	    rdi_dtype |= RDIWatch_WordRead;
	  if (type != 3)
	    rdi_dtype |= RDIWatch_WordWrite;
	  break;

	default:
	  goto bad;
	}
      if ((g.rdi_pointcaps & (rdi_dtype << 2)) != (rdi_dtype << 2))
	return;

      rc = g.rdi_armprocvec->setwatch (g.rdi_armmodule, addr, rdi_type,
				       rdi_dtype, 0, RDI_NoHandle, &h);
    }
  switch (rc)
    {
    case RDIError_NoError:
    case RDIError_NoMorePoints:
      verbose ("Debugpoint set at %x\n", addr);
      break;

    case RDIError_CantSetPoint:
      add_response ("E24", 3);
      verbose ("Too many debugpoints\n");
      return;

    default:
      add_response ("E22", 3);
      verbose ("Unable to set debugpoint\n");
      return;
    }
  if (type == 1)
    {
      uint32_t id;
      int ptype;

      /* Check that we actually got a hardware breakpoint.  */
      rc = g.rdi_armprocvec->info (g.rdi_armmodule, RDIPointStatus_Break,
				   (ARMword *) &id, (ARMword *) &ptype);
      if (rc != RDIError_NoError)
	{
	  /* Assume it's a hardware breakpoint.  */
	  verbose ("Unable to get breakpoint status\n");
	  ptype = 1;
	}
      if (ptype != 1)
	{
	  /* This isn't a hardware breakpoint. Remove it and fail.  */
	  g.rdi_armprocvec->clearbreak (g.rdi_armmodule, h);
	  verbose ("Breakpoint is not a hardware breakpoint\n");
	  add_response ("E24", 3);
	  return;
	}
    }
  if (g.rdi_point_count >= g.rdi_points_allocated)
    {
      if (g.rdi_points_allocated == 0)
	g.rdi_points_allocated = 8;
      else
	g.rdi_points_allocated *= 2;
      g.rdi_points = realloc (g.rdi_points,
			      g.rdi_points_allocated * sizeof (gdb_point));
      if (!g.rdi_points)
	die ("out of memory");
    }
  p = &g.rdi_points[g.rdi_point_count++];
  p->handle = h;
  p->addr = addr;
  p->type = type;
  add_response ("OK", 2);
  return;
bad:
  add_response ("E22", 3);
  verbose ("Bad 'Z' packet\n");
}

/* Respond to a 'z' command.  */
static void
gdb_cmd_rmpoint (const unsigned char * packet, int len)
{
  int type;
  int rc;
  uint32_t addr;
  uint32_t addrlen;
  int i;
  gdb_point *p;

  switch (*packet)
    {
    case '0': /* soft breakpoint */
    case '1': /* hard breakpoint */
    case '2': /* write watchpoint */
    case '3': /* read watchpoint */
    case '4': /* access watchpoint */
      type = *packet - '0';
      break;

    default:
      return;
    }
  packet++;
  if (!(packet && *(packet++) == ','))
    goto bad;
  packet = read_hex (packet, &addr);
  if (!(packet && *(packet++) == ','))
    goto bad;
  packet = read_hex (packet, &addrlen);
  if (!(packet && *(packet++) == '#'))
    goto bad;

  p = g.rdi_points;
  for (i = 0; i < g.rdi_point_count; i++, p++)
    {
      if (p->addr == addr)
	break;
    }
  if (i == g.rdi_point_count)
    goto bad;
  /* TODO: Do we care about the return value. */
  if (type < 2)
    rc = g.rdi_armprocvec->clearbreak (g.rdi_armmodule, p->handle);
  else
    rc = g.rdi_armprocvec->clearwatch (g.rdi_armmodule, p->handle);
  if (rc == RDIError_NoError)
    {
      /* Remove it from the list.  */
      g.rdi_point_count--;
      if (i < g.rdi_point_count)
	*p = g.rdi_points[g.rdi_point_count];
      verbose ("Debugpoint at %x removed\n", addr);
      add_response ("OK", 2);
    }
  else
    {
      verbose ("Unable to debugpoint at %x\n", addr);
      add_response ("E22", 3);
    }
  return;
bad:
  add_response ("E22", 3);
  verbose ("Bad 'Z' packet\n");
}

/* Process a response to a fileio command.  Returns nonzero if this is the
   packet finished the fileio command.  */
static int
process_fileio_packet (const unsigned char * packet, int len)
{
  uint32_t addrlen;
  uint32_t addr;
  unsigned char c;
  unsigned char buff[2];

  c = *(packet++);
  switch (c)
    {
    case 'm':
    case 'M':
      packet = read_hex (packet, &addr);
      if (*(packet++) != ',')
	goto bad;
      packet = read_hex (packet, &addrlen);
      if (addr != 0xffffffff || addrlen != 1)
	goto bad;
      if (c == 'M')
	{
	  if (*(packet++) != ':')
	    goto bad;
	  if (!(isxdigit (packet[0])
		&& isxdigit (packet[1])
		&& packet[2] == '#'))
	    goto bad;
	  g.fileio_data = (from_shex (packet[0]) << 4)
			  | from_shex (packet[1]);
	}
      else
	{
	  if (*packet != '#')
	    goto bad;
	  buff[0] = hexchars[g.fileio_data >> 4];
	  buff[1] = hexchars[g.fileio_data & 0xf];
	  add_response (buff, 2);
	}
      return 0;

    case '#':
    case 'F':
      /* We can't do anything useful if it failed, so just accept
	 any response.  */
      return 1;
    }
bad:
  verbose ("Error while processing FileIO packet\n");
  return 1;
}

/* Process a gdb command packet.  */
static void
process_packet (const unsigned char * packet, int len)
{
  const struct gdb_callback *call;

  for (call = gdb_callbacks; call->c; call++)
    {
      if (call->c == packet[0])
	{
	  call->fn (&packet[1], len - 1);
	  break;
	}
    }
}

/* Acknowledge a gdb packet.  */
static void
send_ack (void)
{
  char c;
  int n;

  c = '+';
  n = send (g.gdb_socket, &c, 1, 0);
  if (n != 1)
    die ("Conection broken");

  verbose ("Send ack\n");
}

/* Request retransmission of a failed packet.  */
static void
send_nack (void)
{
  char c;
  int n;

  c = '-';
  n = send (g.gdb_socket, &c, 1, 0);
  if (n != 1)
    die ("Conection broken");

  verbose ("Requested restransmit\n");
}

/* Read and dispatch gdb packets.  May be called recursively by the hostif
   routines.  */
static void
do_stuff (void)
{
  char buffer[1024];
  int bytes;
  int i;
  unsigned char *packet;
  int c;
  size_t pbuff_size;
  int psize;
  int need_csum;
  unsigned char csum;

  pbuff_size = 4096;
  packet = malloc (pbuff_size);
  if (!packet)
    die ("out of memory");
  psize = 0;
  need_csum = 0;
  for (;;)
    {
      bytes = recv (g.gdb_socket, buffer, sizeof(buffer), 0);
      if (bytes < 0)
	die ("gdb socket read error");
      
      if (bytes == 0)
	break;

      for (i = 0; i < bytes; i++)
	{
	  c = buffer[i];
	  if (g.gdb_response_size)
	    {
	      /* Waiting for an ack.  */
	      if (c == '+')
		{
		  g.gdb_response_size = 0;
		  verbose ("Got ack\n");
		}
	      else if (c == '-')
		{
		  resend_response ();
		  /* Discard any other input.  */
		  i = bytes;
		}
	      /* Assume anything else is a retransmit of the previous command,
		 and ignore it.  */
	      continue;
	    }
	  if (c == '$')
	    {
	      if (psize != 0)
		{
		  verbose ("Abandoning packet after %d bytes '%.*s'\n", psize,
			   psize, packet);
		}
	      psize = 0;
	      need_csum = 0;
	      continue;
	    }

	  if (need_csum)
	    {
	      if (!isxdigit(c))
		{
		  send_nack ();
		  i = bytes;
		  continue;
		}
	      csum = (csum << 4) + from_shex (c);
	      need_csum--;
	      if (!need_csum)
		{
		  if (csum != gdb_checksum (packet, psize))
		    send_nack ();
		  else
		    {
		      send_ack ();
		      clear_response ();
		      if (g.running)
			{
			  /* Response to a fileio packet.  */
			  if (process_fileio_packet (packet, psize))
			    {
			      free (packet);
			      return;
			    }
			}
		      else
			{
			  /* A command packet.  */
			  process_packet (packet, psize);
			}
		      send_response ();
		    }
		  psize = 0;
		  i = bytes;
		}
	      continue;
	    }
	  if (c == '#')
	    {
	      if (packet[0] == 'X')
		verbose ("Got binary X packet\n");
	      else
		verbose ("Got packet: '%.*s'\n", psize, packet);
	      /* End of packet.  */
	      need_csum = 2;
	      csum = 0;
	      /* Copy the end of packet marker to make parsing simpler.  */
	      packet[psize] = c;
	      continue;
	    }

	  /* packet data */
	  /* TODO: X packet escapes.  */
	  /* Make sure teh buffer is big enough to hold this packet and the
	     packet terminator.  */
	  if (psize >= pbuff_size - 1)
	    {
	      packet = realloc(packet, pbuff_size);
	      if (!packet)
		die ("out of memory");
	    }
	  packet[psize++] = c;
	}
    }
  verbose ("gdb connection closed by remote host\n");
  free (packet);
}

/* Display help and exit.  */
static void __attribute__((noreturn))
show_help (void)
{
  printf ("Usage: rdi-stub [-s] [-v] [-q]"
	  " [<address>:]<port> <dso> [<toolconf>]\n");
  printf (" -q     : quiet mode\n");
  printf (" -v     : Verbose (multiple times)\n");
  printf (" -s     : System mode (don't trap exceptions)\n");
  printf (" [<address>]:<port> : Interface and port to listen on"
	  " for gdb connection\n");
  printf (" <dso>  : SORDI .dll/.so to open\n");
  printf (" <toolconf> : toolconf file to open\n");

  exit (1);
}

static void
parse_listenaddr (const char * addr)
{
  char *hostname;
  char *p;
  int i;
  struct hostent *h;

  hostname = strdup (addr);
  p = strchr (hostname, ':');
  if (p && p != hostname)
    {
      *(p++) = 0;
      h = gethostbyname (hostname);
      if (!h)
	die ("Unable to find address %s", hostname);
      if (h->h_addrtype != AF_INET)
	die ("%s is not an ipv4 address\n", hostname);
      g.gdb_listenaddr.sin_addr = *(struct in_addr *) h->h_addr_list[0];
    }
  else
    {
      /* No address specified.  */
      g.gdb_listenaddr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
      if (p)
	p++;
      else
	p = hostname;
    }
  g.gdb_listenaddr.sin_family = AF_INET;
  i = atoi (p);
  if (i == 0)
    die ("Invalid port number: %s", p);
  g.gdb_listenaddr.sin_port = htons (i);
}

int
main(int argc, char ** argv)
{
  const char * dsoname;
  const char * configfile;
  int c;
  
  if (argc == 1)
    show_help ();

  init_sockets ();
  for (;;)
    {
      c = getopt (argc, argv, "chqsv");
      if (c == -1)
	break;

      switch (c)
	{
	case 'v':
	  g.verbose++;
	  break;

	case 'q':
	  g.silent = 1;
	  break;

	case 'c':
	  g.connect = 1;
	  break;

	case 'h':
	case '?':
	  show_help ();

	case ':':
	  /* Should never happen.  */
	  die ("Missing commandline argument");
	  break;
	  
	default:
	  die ("Unknown commandline argument");
	}
    }

  if (optind >= argc)
    die ("Expected [<address>:]port");

  parse_listenaddr (argv[optind++]);

  if (optind >= argc)
    die ("Expected <dsoname>");
  dsoname = argv[optind++];

  if (optind >= argc)
    {
      say ("No toolconf file specified."
	   "This is required for many targets.\n");
      configfile = NULL;
    }
  else
    configfile = argv[optind++];

  if (optind < argc)
    die ("Unexpected commandline argument: %s\n", argv[optind]);

  init_gdb ();
  init_rdi (dsoname, configfile);

  if (g.connect)
    initiate_connection ();
  else
    get_connection ();

  do_stuff ();

  cleanup ();

  return 0;
}
